package adf.sample.module.algorithm;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.information.MessageBuilding;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.DynamicClustering;
import adf.util.WorldUtil;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

public class UFireClustering extends DynamicClustering {

    private int maxDistance;
    private Map<EntityID, Integer> clusterIndexMap;
    private Map<Integer, Set<EntityID>> fireClusters;
    private List<Set<EntityID>> fireClusterList;

    private int updatedTime;

    public UFireClustering(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager) {
        super(ai, wi, si, moduleManager);
        this.init();
    }

    private void init() {
        this.maxDistance = this.scenarioInfo.getFireExtinguishMaxDistance();
        this.clusterIndexMap = new HashMap<>();
        this.fireClusters = new HashMap<>();
        this.fireClusterList = new ArrayList<>();
        for(StandardEntity entity : this.worldInfo.getEntitiesOfType(REFUGE, BUILDING, GAS_STATION, AMBULANCE_CENTRE, FIRE_BRIGADE, POLICE_FORCE)) {
            this.clusterIndexMap.put(entity.getID(), -1);
        }
        this.updatedTime = 0;
    }

    @Override
    public int getClusterNumber() {
        return this.fireClusters.size();
    }

    @Override
    public int getClusterIndex(StandardEntity entity) {
        if(entity instanceof Building) {
            Integer index = this.clusterIndexMap.get(entity.getID());
            if(index != null) { return index; }
        }
        return -1;
    }

    @Override
    public int getClusterIndex(EntityID id) {
        return this.getClusterIndex(this.worldInfo.getEntity(id));
    }

    @Override
    public Collection<StandardEntity> getClusterEntities(int index) {
        Collection<EntityID> cluster = this.getClusterEntityIDs(index);
        return cluster != null ? WorldUtil.convertToEntity(cluster, this.worldInfo) : null;
    }

    @Override
    public Collection<EntityID> getClusterEntityIDs(int index) {
        Set<EntityID> cluster = this.fireClusters.get(index);
        if(cluster != null) { return cluster; }
        if(this.fireClusterList.isEmpty()) {
            this.fireClusterList = new ArrayList<>(this.fireClusters.values());
        }
        if(index >= this.fireClusterList.size()) { return null; }
        return this.fireClusterList.get(index);
    }

    @Override
    public Clustering calc() {
        return this;
    }

    @Override
    public Clustering updateInfo(MessageManager messageManager) {
        // check updated
        int time = this.agentInfo.getTime();
        if(time == this.updatedTime) { return this; }
        // update
        this.updatedTime = time;
        super.updateInfo(messageManager);
        this.fireClusterList.clear();
        // check changed
        Set<EntityID> updated = new HashSet<>();
        for(EntityID id : this.worldInfo.getChanged().getChangedEntities()) {
            StandardEntity entity = this.worldInfo.getEntity(id);
            if(entity instanceof Building) {
                this.update((Building)entity);
                updated.add(id);
            }
        }
        // check MessageBuilding
        messageManager.getReceivedMessageList().stream().filter(message -> message.getClass().equals(MessageBuilding.class)).forEach(message -> {
            MessageBuilding messageBuilding = (MessageBuilding) message;
            if (!updated.contains(messageBuilding.getBuildingID())) {
                int fieryness = messageBuilding.getFieryness();
                if (fieryness > 0 && fieryness < 4) {
                    this.update(WorldUtil.reflectMessage(this.worldInfo, messageBuilding));
                }
            }
        });
        // check MessageFireBrigade (need update Building info)
        messageManager.getReceivedMessageList().stream().filter(message -> message.getClass().equals(MessageFireBrigade.class)).forEach(message -> {
            MessageFireBrigade mfb = (MessageFireBrigade) message;
            int task = mfb.getAction();
            if (task == MessageFireBrigade.ACTION_EXTINGUISH) {
                EntityID target = mfb.getTargetID();
                int index = this.getClusterIndex(target);
                if(index == -1) {
                    this.update((Building)this.worldInfo.getEntity(target));
                }
            }
        });
        return this;
    }

    private void update(Building building) {
        if (!building.isFierynessDefined()) { return; }
        if (!(building.getStandardURN() == REFUGE)) return;

        EntityID buildingID = building.getID();
        int fieryness = building.getFieryness();
        Integer index = this.clusterIndexMap.get(buildingID);
        if (index == null) {
            return;
        }
        Set<EntityID> cluster;
        if(index != -1) {
            cluster = this.fireClusters.get(index);
            if(cluster != null) {
                if (fieryness == 8) {
                    cluster.remove(buildingID);
                    this.clusterIndexMap.put(buildingID, -1);
                } else {
                    if (fieryness > 4) {
                        fieryness -= 4;
                    }
                    int range = (fieryness == 1) ? (this.maxDistance / 5) : (fieryness == 2) ? this.maxDistance : (fieryness == 3) ? (this.maxDistance * 2) : 0;
                    for (StandardEntity entity : this.worldInfo.getObjectsInRange(buildingID, range)) {
                        if (!(entity instanceof Building) || (entity.getStandardURN() == REFUGE)) {
                            continue;
                        }
                        EntityID id = entity.getID();
                        if (id.getValue() == buildingID.getValue()) {
                            continue;
                        }
                        int i = this.clusterIndexMap.get(entity.getID());
                        if (i == -1) {
                            cluster.add(id);
                            this.clusterIndexMap.put(id, index);
                        } else if (index != i) {
                            Set<EntityID> c = this.fireClusters.get(i);
                            if(c != null) {
                                cluster.addAll(c);
                                for (EntityID entityID : c) {
                                    this.clusterIndexMap.put(entityID, index);
                                }
                                this.fireClusters.remove(i);
                            }
                        }
                    }
                }
            }
        } else {
            if(fieryness == 8) { return; }
            if(fieryness > 4) { fieryness -= 4; }
            index = buildingID.getValue();
            cluster = new HashSet<>();
            cluster.add(buildingID);
            int range = (fieryness == 1) ?  (this.maxDistance / 5) : (fieryness == 2) ? this.maxDistance : (fieryness == 3) ?  (this.maxDistance * 2) : 0;
            for (StandardEntity entity : this.worldInfo.getObjectsInRange(buildingID, range)) {
                if (!(entity instanceof Building) || (entity.getStandardURN() == REFUGE)) { continue; }
                EntityID id = entity.getID();
                Building b = (Building) entity;
                if (id.getValue() == buildingID.getValue()) { continue; }
                Integer i = this.clusterIndexMap.get(b.getID());
                if(i == null) {
                    continue;
                }
                if (i == -1) {
                    cluster.add(id);
                    this.clusterIndexMap.put(id, index);
                } else if (index != i) {
                    Set<EntityID> c = this.fireClusters.get(i);
                    if(c != null) {
                        cluster.addAll(c);
                        for (EntityID entityID : c) {
                            this.clusterIndexMap.put(entityID, index);
                        }
                        this.fireClusters.remove(i);
                    }

                }
            }
        }
        this.fireClusters.put(index, cluster);
    }
}