package naito.module.complex;

import adf.agent.communication.MessageManager;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.module.complex.Search;
import adf.util.WorldUtil;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;
import java.util.stream.Collectors;


public class KKScoutSearch extends Search
{
    private int clusterIndex;

    private EntityID targetEntity; // result

    private TreeMap<Integer, Integer> areaInformation; // EntityID -> Time
    private TreeMap<Integer, Collection<EntityID>> updatedInformation; // Time -> List<EntityID>

    private final int searchRange; // range of searching
    private final int threshold; // life of information

    private int helpCount;
    private int limits;
    private final int limitsToSearch;
    private List<StandardEntityURN> searchURN;

    public KKScoutSearch(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager) {
        super(ai, wi, si, moduleManager);
        this.clusterIndex = -1;
        this.targetEntity = null;
        this.searchRange = 4;
        this.threshold = 50;
        this.limits = 0;
        this.limitsToSearch = 4;

        searchURN = Arrays.asList(StandardEntityURN.BUILDING, StandardEntityURN.ROAD);
        this.init();
    }

    private boolean isConnector(Area area)
    {
        List<EntityID> neighbours = area.getNeighbours();
        return (neighbours.size() < 3) &&
                neighbours.stream()
                        .anyMatch((a) -> searchURN.contains(worldInfo.getEntity(a).getStandardURN()));
    }

    private void init() {
        /*
         * construct area information map
         *  and construct updated information map
         */
        this.areaInformation = new TreeMap<>();
        List<EntityID> up = new ArrayList<>();
        Area r;
        for (StandardEntity next : this.worldInfo) {
            // look at map and select "search" area
            if(searchURN.contains(next.getStandardURN())) {
                r = (Area)next;
                areaInformation.put(next.getID().getValue(), agentInfo.getTime());
                up.add(next.getID());
            }
        }

        this.updatedInformation = new TreeMap<>();
        updatedInformation.put(agentInfo.getTime(), up);

    }

    @Override public Search updateInfo(MessageManager messageManager) {
        /*
         * update information
         * */
        int previousTime;
        int currentTime = agentInfo.getTime();
        helpCount = messageManager.getHeardAgentHelpCount();

        // updated in this time
        List<EntityID> updated = new ArrayList<>();
        Collection<StandardEntity> changes =
                WorldUtil.convertToEntity(this.worldInfo.getChanged().getChangedEntities(), this.worldInfo);
        // search all changes
        for(StandardEntity next : changes) {
            if(searchURN.contains(next.getStandardURN())) {
                // to update area Information
                if(areaInformation.containsKey(next.getID().getValue()))
                {
                    previousTime = this.areaInformation.get(next.getID().getValue());
                    this.areaInformation.put(next.getID().getValue(), currentTime);

                    // move 'Area' to updated
                    updatedInformation.get(previousTime).remove(next.getID());
                    if(this.updatedInformation.get(previousTime).isEmpty()) {
                        updatedInformation.remove(previousTime);
                    }
                    updated.add(next.getID());
                }
            }
            if(this.targetEntity != null
                    && (this.targetEntity.getValue() == next.getID().getValue()
                    || (limits--) > 0))

            {
                this.targetEntity = null;
            }
        }
        if(!updated.isEmpty()) {
            this.updatedInformation.put(currentTime, updated);
        }

        return this;
    }


    private int countNeighboursBlockadeRepairCosts(EntityID e)
    {
        if (!searchURN.contains(worldInfo.getEntity(e).getStandardURN())) {
            return 0;
        }
        Collection<EntityID> neighbours = ((Area)worldInfo.getEntity(e)).getNeighbours();
        Road road;
        int blockade = 0;
        for(EntityID next : neighbours) {
            if(this.worldInfo.getEntity(next).getStandardURN() == StandardEntityURN.ROAD)
            {
                road = (Road)this.worldInfo.getEntity(next);
                if(road.isBlockadesDefined() && !road.getBlockades().isEmpty()) {
                    blockade += road.getBlockades()
                            .stream()
                            .map((a) -> (Blockade)this.worldInfo.getEntity(a))
                            .filter(Blockade::isRepairCostDefined)
                            .map(Blockade::getRepairCost)
                            .reduce((a, b) -> a + b)
                            .orElse(0);
                }
            }
        }

        return blockade;
    }

    private boolean considerCivilianThreatenInBuilding(Building building)
    {
        Collection<EntityID> neighbours = building.getNeighbours();
        int broken = (building.isBrokennessDefined()) ? building.getBrokenness() : 0;
        int blockade = countNeighboursBlockadeRepairCosts(building.getID());
        int temp = (building.isTemperatureDefined()) ? building.getTemperature() : 0;
        int fiery = (building.isFierynessDefined()) ? building.getFieryness() : 0;

        boolean result = (broken > 0) || (blockade > 0) || (temp/100 > 3) || (fiery != 0);

        return result;
    }

    private int evaluateTarget(EntityID a, EntityID b)
    {
        EntityID[] ids = {a, b};
        int[] cost = {0, 0};

        for (int i = 0; i < 2; i++) {
            // distance, and blockade costs
            cost[i] = worldInfo.getDistance(this.agentInfo.getID(), ids[i]);
            cost[i] += countNeighboursBlockadeRepairCosts(ids[i]);
            cost[i] += (isConnector((Area)worldInfo.getEntity(ids[i]))) ? 0 : 10000;


            // add burning costs
            if(worldInfo.getEntity(ids[i]).getStandardURN() == StandardEntityURN.BUILDING) {
                Building building = (Building)worldInfo.getEntity(ids[i]);
                if(building.isFierynessDefined() && ((building.getFieryness() & 0x03) != 0))
                {
                    // burning when temp > 400
                    cost[i] += (building.isTemperatureDefined()) ? building.getTemperature() : 400;
                }
                cost[i] += building.getNeighbours()
                        .stream()
                        .map((n) -> worldInfo.getEntity(n))
                        .filter(n -> n.getStandardURN()==StandardEntityURN.BUILDING)
                        .map(n -> (Building)n)
                        .filter(Building::isTemperatureDefined)
                        .map(Building::getTemperature)
                        .reduce((n, m) -> n + m)
                        .orElse(0);
            }
            else if(worldInfo.getEntity(ids[i]).getStandardURN() == StandardEntityURN.REFUGE)
            {
                cost[i] += 10000;
            }
        }

        // compare
        return cost[0] - cost[1];
    }


    @Override public Search calc() {
        Collection<EntityID> dest = null; // destinations

        // searching with received Civilians Help
        if (helpCount > 0) {
            // searching victim
            Collection<StandardEntity> targets
                    = this.worldInfo.getObjectsInRange(agentInfo.getPosition(),
                    this.scenarioInfo.getPerceptionLosMaxDistance());

            int nearCiv = (int)targets.stream()
                    .filter((a) -> a.getStandardURN() == StandardEntityURN.CIVILIAN)
                    .map((a) -> (Civilian)a)
                    .filter((a) -> a.isBuriednessDefined() && a.getBuriedness() > 0)
                    .count();

            if(helpCount - nearCiv > 0) {
                // searching "civilian possible" building
                Building building;
                HashSet<EntityID> deemingBuilding = new HashSet<>();
                for (StandardEntity next : targets) {
                    if (next.getStandardURN() == StandardEntityURN.BUILDING) {
                        building = (Building) next;
                        // evaluate
                        if (considerCivilianThreatenInBuilding(building)) {
                            // collection ROAD Entity
                            List<EntityID> collect = building.getNeighbours().stream()
                                    .filter(a -> worldInfo.getEntity(a).getStandardURN()
                                            == StandardEntityURN.ROAD)
                                    .collect(Collectors.toList());
                            deemingBuilding.addAll(collect);
                        }
                    } // fi (next is BUILDING)
                } // rof

                // decide go to any building possible
                if (!deemingBuilding.isEmpty()) {
                    dest = deemingBuilding;
                }
            } // fi (helpCount - nearCiv > 0)
        } // fi (helpCount available)

        // searching normally
        if(this.targetEntity != null
                && this.agentInfo.getPosition().getValue() != this.targetEntity.getValue()
                && (limits--) < 0)
        {
            // still, use previous calc result.
            dest = null;
        }
        else if(dest == null) {
            // searching road
            int range = this.scenarioInfo.getPerceptionLosMaxDistance() * this.searchRange;
            int oldestTime = this.updatedInformation.firstKey();

            // Area Searching
            Collection<EntityID> targetAreas
                    = this.worldInfo.getObjectIDsInRange(this.agentInfo.me(), range);
            // remove current area
            targetAreas.remove(this.agentInfo.getPosition());

            // collect viewing area of Information target of Searching
            int timeRange = oldestTime + this.threshold;
            if(agentInfo.getTime() < timeRange) {
                timeRange = oldestTime;
            }
            Map<Integer, Collection<EntityID>> viewing
                    = this.updatedInformation.headMap(timeRange, true);

            // is oldest information in Viewing?
            Set<EntityID> searchingEntities = new HashSet<>(targetAreas);
            viewing.values().forEach(searchingEntities::retainAll);

            // if empty, target is oldest of having
            if(!searchingEntities.isEmpty()) {
                dest = searchingEntities;
            }
            else{
                dest = this.updatedInformation.firstEntry().getValue();
            }
        }

        // get destination
        if(dest != null && !dest.isEmpty()) {
            // decide the best place to go
            // TODO: more good function, current way is comparing

            this.targetEntity = dest.stream()
                    .min(this::evaluateTarget)
                    .get();
            if(this.targetEntity != null)
                this.limits = this.limitsToSearch;
        }

        return this;
    }




    @Override public EntityID getTarget() {
        return this.targetEntity;
    }

    @Override public Search precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        return this;
    }

    @Override public Search resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        return this;
    }

    @Override public Search preparate() {
        super.preparate();
        return this;
    }

}


