package naito.tactics;

import adf.agent.action.Action;
import adf.agent.action.ambulance.ActionRescue;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageBuilding;
import adf.agent.communication.standard.bundle.information.MessageCivilian;
import adf.agent.communication.standard.bundle.information.MessageRoad;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.complex.HumanSelector;
import adf.component.module.complex.Search;
import adf.component.tactics.TacticsAmbulance;
import naito.SampleModuleKey;
import adf.util.WorldUtil;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.GeometryTools2D;
import java.util.*;


public class KKAmbulance extends TacticsAmbulance
{
    private PathPlanning pathPlanning;
    private Clustering clustering;
    private HumanSelector humanSelector;
    private Search search;

    private List<Point2D> positionHistory;
    private int clearRequestCount;
    private Action lastAction;

    @Override
    public void initialize(
        AgentInfo      agentInfo,
        WorldInfo      worldInfo,
        ScenarioInfo   scenarioInfo,
        ModuleManager  moduleManager,
        MessageManager messageManager
    ){

        moduleManager.getExtAction(
            SampleModuleKey.AMBULANCE_ACTION_TRANSPORT,
            "naito.extaction.ActionTransport"
        );
        moduleManager.getExtAction(
            SampleModuleKey.AMBULANCE_ACTION_SEARCH,
            "naito.extaction.ActionSearch"
        );

        this.positionHistory = new LinkedList<>();
        clearRequestCount = 0;
    }

    @Override
    public void precompute(
        AgentInfo      agentInfo,
        WorldInfo      worldInfo,
        ScenarioInfo   scenarioInfo,
        ModuleManager  moduleManager,
        PrecomputeData precomputeData
    ){
        pathPlanning = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_PATH_PLANNING,
            "naito.module.algorithm.SamplePathPlanning"
        );
        clustering = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_CLUSTERING,
            "naito.module.algorithm.SampleKMeans"
        );
        humanSelector = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_HUMAN_SELECTOR,
            "naito.module.complex.SampleVictimSelector"
        );
        search = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_SEARCH,
            "naito.module.complex.SampleSearch"
        );

        pathPlanning.precompute(precomputeData);
        clustering.precompute(precomputeData);
        humanSelector.precompute(precomputeData);
        search.precompute(precomputeData);
    }

    @Override
    public void resume(
        AgentInfo      agentInfo,
        WorldInfo      worldInfo,
        ScenarioInfo   scenarioInfo,
        ModuleManager  moduleManager,
        PrecomputeData precomputeData
    ){
        pathPlanning = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_PATH_PLANNING,
            "naito.module.algorithm.SamplePathPlanning"
        );
        clustering = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_CLUSTERING,
            "naito.module.algorithm.SampleKMeans"
        );
        humanSelector = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_HUMAN_SELECTOR,
            "naito.module.complex.SampleVictimSelector"
        );
        search = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_SEARCH,
            "naito.module.complex.SampleSearch"
        );

        pathPlanning.resume(precomputeData);
        clustering.resume(precomputeData);
        humanSelector.resume(precomputeData);
        search.resume(precomputeData);
    }

    @Override
    public void preparate(
        AgentInfo     agentInfo,
        WorldInfo     worldInfo,
        ScenarioInfo  scenarioInfo,
        ModuleManager moduleManager
    ){
        pathPlanning = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_PATH_PLANNING,
            "naito.module.algorithm.SamplePathPlanning"
        );
        clustering = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_CLUSTERING,
            "naito.module.algorithm.SampleKMeans"
        );
        humanSelector = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_HUMAN_SELECTOR,
            "naito.module.complex.SampleVictimSelector"
        );
        search = moduleManager.getModule(
            SampleModuleKey.AMBULANCE_MODULE_SEARCH,
            "naito.module.complex.SampleSearch"
        );

        pathPlanning.preparate();
        clustering.preparate();
        humanSelector.preparate();
        search.preparate();
    }

    private void updateInfo(WorldInfo worldInfo, MessageManager messageManager)
    {
        List<CommunicationMessage> messages = messageManager.getReceivedMessageList();

        for (CommunicationMessage message : messages)
        {
            if (message instanceof MessageAmbulanceTeam)
                WorldUtil.reflectMessage(worldInfo, (MessageAmbulanceTeam)message);

            if (message instanceof MessageCivilian)
                WorldUtil.reflectMessage(worldInfo, (MessageCivilian)message);

            if (message instanceof MessageBuilding)
                WorldUtil.reflectMessage(worldInfo, (MessageBuilding)message);

            if (message instanceof MessageRoad)
                WorldUtil.reflectMessage(worldInfo, (MessageRoad)message);
        }
    }

    private void sendMessage(
        AgentInfo      agentInfo,
        WorldInfo      worldInfo,
        MessageManager messageManager,
        Action         action
    ){
        if (action == null) return;

        if ((action instanceof ActionMove) && agentInfo.someoneOnBoard() != null)
        {
            ActionMove actionMove = (ActionMove)action;
            messageManager.addMessage(
                new MessageCivilian(true, (Civilian)agentInfo.someoneOnBoard())
            );

            messageManager.addMessage(
                new MessageAmbulanceTeam(
                    true,
                    (AmbulanceTeam)agentInfo.me(),
                    MessageAmbulanceTeam.ACTION_MOVE,
                    actionMove.getPath().get(actionMove.getPath().size()-1)
                )
            );
        }

        if (action instanceof ActionRescue)
        {
            ActionRescue actionRescue = (ActionRescue)action;
            messageManager.addMessage(
                new MessageAmbulanceTeam(
                    true,
                    (AmbulanceTeam)agentInfo.me(),
                    MessageAmbulanceTeam.ACTION_RESCUE,
                    actionRescue.getTarget()
                )
            );
        }
    }

    @Override
    public Action think(
        AgentInfo agentInfo,
        WorldInfo worldInfo,
        ScenarioInfo scenarioInfo,
        ModuleManager moduleManager,
        MessageManager messageManager
    ){
        Action action = null;

        this.pathPlanning.updateInfo(messageManager);
        this.clustering.updateInfo(messageManager);
        this.humanSelector.updateInfo(messageManager);
        this.search.updateInfo(messageManager);

        this.updateInfo(worldInfo, messageManager);

        AmbulanceTeam me = (AmbulanceTeam)agentInfo.me();

        //check buriedness
        if (me.isBuriednessDefined() && me.getBuriedness() > 0)
        {
            messageManager.addMessage(
                new MessageAmbulanceTeam(true, me, MessageAmbulanceTeam.ACTION_REST, me.getPosition())
            );
            this.lastAction = new ActionRest();
            return new ActionRest();
        }

        if (agentInfo.getTime() <= scenarioInfo.getKernelAgentsIgnoreuntil())
        {
            this.lastAction = new ActionRest();
            return new ActionRest();
        }

        this.positionHistory.add(new Point2D(agentInfo.getX(), agentInfo.getY()));
        if (this.positionHistory.size() > 3)
            this.positionHistory.remove(0);


        if (this.isAgentStop() && this.clearRequestCount == 0)
        {
            this.clearRequestCount =
                (this.isAgentInBlockade(agentInfo, worldInfo)) ? 2 : 1;
        }

        if (this.clearRequestCount > 0)
        {
            this.clearRequestCount--;
            messageManager.addMessage(
                new MessageAmbulanceTeam(true, me, MessageAmbulanceTeam.ACTION_REST, me.getPosition())
            );

            if (this.clearRequestCount <= 0)
                this.clearRequestCount = -1;
        }


        // Transporting
        Human injured = agentInfo.someoneOnBoard();
        if (injured != null)
        {
            ExtAction extActionTransport = moduleManager.getExtAction(
                SampleModuleKey.AMBULANCE_ACTION_TRANSPORT,
                "naito.extaction.ActionTransport"
            );
            action = extActionTransport.setTarget(injured.getID()).calc().getAction();
        }

        // Searched, Rescue
        if (action == null)
        {
            EntityID target = this.humanSelector.calc().getTarget();
            if (target != null)
            {
                ExtAction extActionTransport = moduleManager.getExtAction(
                    SampleModuleKey.AMBULANCE_ACTION_TRANSPORT,
                    "naito.extaction.ActionTransport"
                );

                action = extActionTransport.setTarget(target).calc().getAction();
            }
        }

        // Search
        if (action == null)
        {
            // Nothing to do
            EntityID target = this.search.calc().getTarget();
            ExtAction extActionSearch = moduleManager.getExtAction(
                SampleModuleKey.AMBULANCE_ACTION_SEARCH,
                "naito.extaction.ActionSearch"
            );
            action = extActionSearch.setTarget(target).calc().getAction();
        }

        //this.sendMessage(agentInfo, worldInfo, messageManager, action);

        this.lastAction = action;
        return action;
    }

    private final static int AGENT_SIZE = 500;

    private boolean isAgentStop()
    {
        if (this.positionHistory.size() < 3)
            return false;

        Point2D pos1 = this.positionHistory.get(this.positionHistory.size()-1);
        Point2D pos2 = this.positionHistory.get(this.positionHistory.size()-2);
        Point2D pos3 = this.positionHistory.get(this.positionHistory.size()-3);
        if (GeometryTools2D.getDistance(pos1, pos2) > AGENT_SIZE * 2)
        {
            this.clearRequestCount = 0;
            return false;
        }

        if (GeometryTools2D.getDistance(pos2, pos3) > AGENT_SIZE * 2)
        {
            this.clearRequestCount = 0;
            return false;
        }

        if (this.lastAction != null &&
            this.lastAction instanceof ActionRescue
        ){
            this.clearRequestCount = 0;
            return false;
        }

        return true;
    }

    private boolean isAgentInBlockade(AgentInfo agentInfo, WorldInfo worldInfo)
    {
        StandardEntity position = worldInfo.getEntity(agentInfo.getPosition());
        if (!(position instanceof Area))
            return false;

        Area positionArea = (Area)position;

        if (!positionArea.isBlockadesDefined())
            return false;

        for (EntityID blockadeID : positionArea.getBlockades())
        {
            StandardEntity entity = worldInfo.getEntity(blockadeID);
            if (!(entity instanceof Blockade))
                continue;

            Blockade blockade = (Blockade)entity;
            if (blockade.getShape().contains(agentInfo.getX(), agentInfo.getY()))
                return true;
        }

        return false;
    }
}
