package naito.util.passable;

import naito.util.passable.astar.AStar;
import rescuecore2.misc.geometry.*;
import rescuecore2.standard.entities.Area; import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Blockade;
import adf.agent.info.WorldInfo;
import naito.util.passable.astar.Index;

import java.util.*;
import java.util.stream.*;

public class CheckAreaPassable
{
    private static final double AGENT_SIZE = 500;

    public boolean check(Area area, Line2D in, Line2D out, WorldInfo wi)
    {
        //  @STEP1 make grid of area
        Point2D[][] grid = AreaGrid.create(area, in, out);

        //  @STEP2 remove points of grid in area's blockades
        List<Blockade> blockades = AreaUtil.toBlockades(area, wi);
        this.removePointInShapes(
            grid,
            AreaUtil.toShapes(blockades)
        );


        //  @STEP3 astar for grid of area
        AStar<Point2D> astar = new AStar<>(grid);
        //astar.print();

        //  astar : set begin and goal index
        List<Index> beginIndexes = this.findIndexesOfContiguous(in , grid);
        List<Index>  goalIndexes = this.findIndexesOfContiguous(out, grid);
        if (beginIndexes.size() == 0) return false;
        if ( goalIndexes.size() == 0) return false;

        astar.setBegin(
            this.extractIndexOfClosest(GeometryMath.center(in ), beginIndexes, grid)
        );
        astar.setGoal(
            this.extractIndexOfClosest(GeometryMath.center(out),  goalIndexes, grid)
        );

        List<Point2D> path = astar.calc().getResult();
        if (path.size() == 0) return false;


        //  @STEP4 once check path is wrong or not
        boolean hasProblem = this.hasCrossing(
            GeometryTools2D.pointsToLines(path),
            AreaUtil.toLines(blockades)
        );

        //System.out.println("CheckAreaPassable:" + area.getID() + "=" + !hasProblem);

        return !hasProblem;
    }

    public boolean check(Area area, Area in, Area out, WorldInfo wi)
    {
        Edge  inEdge = this.findDuplicateEdge(area, in);
        if ( inEdge == null) return false; 
        Edge outEdge = this.findDuplicateEdge(area, out);
        if (outEdge == null) return false;

        return
            this.check(area, inEdge.getLine(), outEdge.getLine(), wi);
    }

    public boolean check(Area area, Point2D inPoint, Area out, WorldInfo wi)
    {
        Edge outEdge = this.findDuplicateEdge(area, out);
        if (outEdge == null) return false;

        return
            this.check(area, inPoint, outEdge.getLine(), wi);
    }

    public boolean check(Area area, Point2D inPoint, Line2D out, WorldInfo wi)
    {
        //  @STEP1 make grid of area
        Point2D[][] grid = AreaGrid.create(area, inPoint, out);

        Index beginIndex = this.extractIndexOfClosest(inPoint, grid);

        //  @STEP2 remove points of grid in area's blockades
        List<Blockade> blockades = AreaUtil.toBlockades(area, wi);
        this.removePointInShapes(
            grid,
            AreaUtil.toShapes(blockades)
        );


        //  @STEP3 astar for grid of area
        AStar<Point2D> astar = new AStar<>(grid);

        //  astar : set begin and goal index
        List<Index>  goalIndexes = this.findIndexesOfContiguous(out, grid);
        if ( goalIndexes.size() == 0) return false;

        astar.setBegin(beginIndex);
        astar.setGoal(
            this.extractIndexOfClosest(GeometryMath.center(out),  goalIndexes, grid)
        );

        List<Point2D> path = astar.calc().getResult();
        if (path.size() == 0) return false;


        //  @STEP4 once check path is wrong or not
        boolean hasProblem = this.hasCrossing(
            GeometryTools2D.pointsToLines(path),
            AreaUtil.toLines(blockades)
        );

        //System.out.println("CheckAreaPassable:" + area.getID() + "=" + !hasProblem);

        return !hasProblem;
    }

    public boolean check(Area area, Area in, Point2D outPoint, WorldInfo wi)
    {
        Edge inEdge = this.findDuplicateEdge(area, in);
        if (inEdge == null) return false;

        return
            this.check(area, inEdge.getLine(), outPoint, wi);
    }

    public boolean check(Area area, Line2D in, Point2D outPoint, WorldInfo wi)
    {
        //  @STEP1 make grid of area
        Point2D[][] grid = AreaGrid.create(area, in, outPoint);

        Index goalIndex = this.extractIndexOfClosest(outPoint, grid);

        //  @STEP2 remove points of grid in area's blockades
        List<Blockade> blockades = AreaUtil.toBlockades(area, wi);
        this.removePointInShapes(
            grid,
            AreaUtil.toShapes(blockades)
        );


        //  @STEP3 astar for grid of area
        AStar<Point2D> astar = new AStar<>(grid);

        //  astar : set begin and goal index
        List<Index>  beginIndexes = this.findIndexesOfContiguous(in, grid);
        if (beginIndexes.size() == 0) return false;

        astar.setBegin(
            this.extractIndexOfClosest(GeometryMath.center(in),  beginIndexes, grid)
        );
        astar.setGoal(goalIndex);

        List<Point2D> path = astar.calc().getResult();
        if (path.size() == 0) return false;


        //  @STEP4 once check path is wrong or not
        boolean hasProblem = this.hasCrossing(
            GeometryTools2D.pointsToLines(path),
            AreaUtil.toLines(blockades)
        );

        //System.out.println("CheckAreaPassable:" + area.getID() + "=" + !hasProblem);

        return !hasProblem;
    }

    private Edge findDuplicateEdge(Area area1, Area area2)
    {
        List<Edge> edges1 = area1.getEdges().stream()
            .filter(Edge::isPassable)
            .collect(Collectors.toList());

        List<Edge> edges2 = area2.getEdges().stream()
            .filter(Edge::isPassable)
            .collect(Collectors.toList());

        for (Edge edge1 : edges1) for (Edge edge2 : edges2)
        {
            Line2D line1 = edge1.getLine();
            Line2D line2 = edge2.getLine();
            if (this.isEqual(line1, line2))
                return edge1;
        }

        return null;
    }

    private boolean isEqual(Line2D line1, Line2D line2)
    {
        Point2D origin1 = line1.getOrigin();
        Point2D origin2 = line2.getOrigin();
        Point2D end1 = line1.getEndPoint();
        Point2D end2 = line2.getEndPoint();
        if (origin1.equals(origin2) && end1.equals(end2))
            return true;

        if (origin1.equals(end2) && origin2.equals(end1))
            return true;

        return false;
    }

    private void removePointInShapes(
        Point2D[][]          points,
        List<java.awt.Shape> shapes
    ){
        int maxX = points.length;
        int maxY = points[0].length;

        for (int x=0; x<maxX; x++) for (int y=0; y<maxY; y++)
        {
            Point2D point = points[x][y];
            if (point == null) continue;

            for (java.awt.Shape shape : shapes)
                if ( shape.contains(point.getX(), point.getY()) )
                {
                    points[x][y] = null;
                    break;
                }
        }

        return;
    }

    private List<Index> findIndexesOfContiguous(
        Line2D      line,
        Point2D[][] points
    ){
        int maxX = points.length;
        int maxY = points[0].length;

        List<Index> indexes = new ArrayList<>();

        for (int x=0; x<maxX; x++) for (int y=0; y<maxY; y++)
        {
            Point2D point = points[x][y];
            if (point == null) continue;

            //if (GeometryMath.distance(line, point) < AGENT_SIZE*1.5)
            if (GeometryMath.distance(line, point) < AGENT_SIZE*1.5)
                indexes.add( new Index(x, y) );
        }

        return indexes;
    }

    private Index extractIndexOfClosest(
        Point2D     point,
        List<Index> candidacies,
        Point2D[][] points
    ){
        Index  bestIndex   = null;
        double minDistance = Double.MAX_VALUE;

        for (Index index: candidacies)
        {
            double distance = GeometryTools2D.getDistance(
                point,
                index.toElement(points)
            );

            if (distance < minDistance)
            {
                bestIndex   = index;
                minDistance = distance;
            }
        }

        return bestIndex;
    }

    private Index extractIndexOfClosest(
        Point2D     point,
        Point2D[][] points
    ){
        int maxX = points.length;
        int maxY = points[0].length;

        Index  bestIndex   = null;
        double minDistance = Double.MAX_VALUE;

        for (int x=0; x<maxX; x++) for (int y=0; y<maxY; y++)
        {
            if (points[x][y] == null)
                continue;

            double distance = GeometryTools2D.getDistance(
                point,
                points[x][y]
            );

            if (distance < minDistance)
            {
                bestIndex   = new Index(x, y);
                minDistance = distance;
            }
        }

        return bestIndex;
    }

    private boolean hasCrossing(List<Line2D> lines1, List<Line2D> lines2)
    {
        for (Line2D line1 : lines1) for (Line2D line2 : lines2)
        {
            Point2D intersection =
                GeometryTools2D.getSegmentIntersectionPoint(line1, line2);

            if (intersection != null)
                return true;
        }

        return false;
    }
}
