package naito.util.passable.astar;

import java.util.*;
import java.util.stream.*;

public class AStar<T>
{
    private naito.util.passable.astar.Node<T>[][] nodes;
    private Index focused = null;
    private Index goal    = null;

    // for debug
    private List<naito.util.passable.astar.Node<T>> nodePath;
    private List<Index> pathIndex;

    public AStar(T[][] contents)
    {
        this.nodes = this.contentsToNodes(contents);
    }

    private naito.util.passable.astar.Node<T>[][] contentsToNodes(T[][] contents)
    {
        int maxX = contents.length;
        int maxY = contents[0].length;

        @SuppressWarnings({"rawtypes","unchecked"})
        naito.util.passable.astar.Node<T>[][] nodes = new naito.util.passable.astar.Node[maxX][maxY];

        for (int x=0; x<maxX; x++)
            for (int y=0; y<maxY; y++)
            {
                if (contents[x][y] == null) continue;

                nodes[x][y] = new naito.util.passable.astar.Node<T>(contents[x][y]);
            }

        return nodes;
    }

    public AStar<T> calc()
    {
        //this.print();
        this.nodePath = new ArrayList<>();
        if (this.focused == null) return this;
        if (this.goal    == null) return this;
        if (this.focused.toElement(this.nodes) == null) return this;
        if (this.goal.toElement(this.nodes) == null) return this;

        while (this.focused != null &&
              !this.focused.equals(this.goal)
        ){
            this.toNode(this.focused,  this.nodes).setState(State.Close);
            this.open(  this.focused,  this.nodes);
            this.focused = this.select(this.nodes);
        }

        //return this.trace(this.focused, this.nodes);
        this.nodePath = this.trace(this.focused, this.nodes);
        //this.print();

        return this;
    }

    private void open(Index center, naito.util.passable.astar.Node<T>[][] nodes)
    {
        int maxX = nodes.length;
        int maxY = nodes[0].length;
        naito.util.passable.astar.Node<T> centerNode = this.toNode(center, nodes);

        for (int dx : new int[]{-1, 0, 1})
            for (int dy : new int[]{-1, 0, 1})
            {
                int x = center.x + dx;
                int y = center.y + dy;

                if (dx == 0 && dy == 0) continue;
                if (x < 0)              continue;
                if (y < 0)              continue;
                if (x >= maxX)          continue;
                if (y >= maxY)          continue;

                naito.util.passable.astar.Node<T> node = nodes[x][y];
                if (node == null)                  continue;
                if (node.getState() != State.None) continue;

                node.setState(State.Open);
                node.setFootpoint(center);
                node.setCostToSelf( centerNode.getCostToSelf()+centerNode.getStep() );
                node.setCostToGoal( Math.abs(this.goal.x-x), Math.abs(this.goal.y-y) );
            }

        return;
    }

    private Index select(naito.util.passable.astar.Node<T>[][] nodes)
    {
        int maxX = nodes.length;
        int maxY = nodes[0].length;

        //  find node of min score
        Index   minIndex = null;
        naito.util.passable.astar.Node<T> minNode  = null;

        for (int x=0; x<maxX; x++)
            for (int y=0; y<maxY; y++)
            {
                naito.util.passable.astar.Node<T> node = this.nodes[x][y];
                if (node == null)
                    continue;
                if (node.getState() != State.Open)
                    continue;
                if (minIndex != null && !minNode.isCloser(node))
                    continue;

                minIndex = new Index(x, y);
                minNode  = node;
            }

        return minIndex;
    }

    private List<naito.util.passable.astar.Node<T>> trace(Index endindex, naito.util.passable.astar.Node<T>[][] nodes)
    {
        List<naito.util.passable.astar.Node<T>> nodePath = new ArrayList<>();
        Index footpoint = endindex;

        this.pathIndex = new ArrayList<>();

        while (footpoint != null)
        {
            naito.util.passable.astar.Node<T> node = this.toNode(footpoint, this.nodes);
            nodePath.add(node);
            this.pathIndex.add(footpoint);
            footpoint = node.getFootpoint();
        }

        Collections.reverse(nodePath);
        return nodePath;
    }

    private naito.util.passable.astar.Node<T> toNode(Index index, naito.util.passable.astar.Node<T>[][] nodes)
    {
        return nodes[index.x][index.y];
    }

    public void setBegin(Index index)
    {
        this.focused= index;
    }

    public void setGoal(Index index)
    {
        this.goal = index;
    }

    public naito.util.passable.astar.Node<T>[][] getNodes()
    {
        return this.nodes;
    }

    public List<T> getResult()
    {
        return this.nodePath.stream()
            .map(naito.util.passable.astar.Node<T>::getContent)
            .collect(Collectors.toList());
    }

    public List<naito.util.passable.astar.Node<T>> getResultNodes()
    {
        return this.nodePath;
    }

    public void print()
    {
        int maxX = this.nodes.length;
        int maxY = this.nodes[0].length;

        for (int y=maxY-1; y>=0; y--)
        {
            for (int x=0; x<maxX; x++)
            {
                System.out.print( this.toMark( new Index(x, y) ) + " ");
            }

            System.out.println("");
        }

        System.out.println("===============================================");
    }

    private char toMark(Index index)
    {
        if ( this.focused != null && this.focused.equals(index) ) return 'f';
        if ( this.goal != null && this.goal.equals(index) )    return 'g';

        naito.util.passable.astar.Node<T> node = this.toNode(index, this.nodes);
        if (node == null) return '-';
        if ( this.nodePath != null && this.pathIndex.contains(index) && node.getStep() == 2)  return 'b';
        if (node.getStep() == 2) return '-';
        if ( this.nodePath != null && this.pathIndex.contains(index) )  return 'p';

        switch ( node.getState() )
        {
            case Open : return 'o';
            case Close: return 'c';
            case None : return 'n';
        }

        return '-';
    }

}
